import { CardData } from "./Models/CardData";
import { ClientMessage, ClientTechnicalProblemDetectedMessage, ClientChallengeRequiredMessage, ClientFingerPrintRequiredMessage, ClientAuthenticatedMessage } from "./Models/ClientMessage";
import { HostMessage } from "./Models/HostMessage";
import { ModelData } from "./Models/ModelData";

export class Engine { 

    /**
     * Base url for Api
     */
    apiBaseUrl : string;

    /**
     * The session token as generated by ePayTools
     */
    private sessionToken!: string;

    /**
     * handle for a Polling background task
     */
    private backgroundPollingInterval!: number;

    /**
     * handle for the timer on the Polling operation
     */
    private timerPollingHandle!: number;

    /**
     * An indication for the polling task to stop
     */
    private stopPollingSemaphore: boolean = false;

    /**
     * iFrame for challenge/fingerprint
     */
    private iFrame!: HTMLIFrameElement;
    
    /**
     * Resolver of entry point
     */
    private resolver!: (msg: ClientMessage) => void;

/**
 * Constructor
 * @param apiBaseUrl The url of the server. Leave the default for production
 */
    constructor(apiBaseUrl: string = "https://api.epaytools.com") {
        this.apiBaseUrl = apiBaseUrl;
    }

    /**
     * Main function to perform a 3d secure authentication session
     * 
     * @param {string} sessionToken - token obtained from the Orchestra service
     * @param {string} iFrameElementSelector - Selector of element in which a challenge iFrame is to be created
     * @param {CardData} cardData - The card details
     */
     async perform3ds  (sessionToken: string, iFrameElementSelector: string, cardData: CardData): Promise<ClientMessage> {
        if (!iFrameElementSelector || !cardData  || !sessionToken) {
            const err = "mainContentId || iFrameElementId or cardData or sessionToken are empty; quitting";
            console.error(err);
            return Promise.reject(err);
        }

        const iFrame = document.querySelector(iFrameElementSelector);
        if (!iFrame) {
            const err = iFrameElementSelector + " not accessible";
            console.error(err);
            return Promise.reject(err);
        }
        // Store iFrame element
        this.iFrame = this.forceCast(iFrame)

        // store  sessionToken
        this.sessionToken = sessionToken;
        // Init 3DS session 
        this.init3DSecAuthSession(cardData);
                
        return new Promise<ClientMessage>((resolve, reject) => {
            this.resolver = resolve;
            // rejector = reject;
        });
    }

    /**
     * init 3DS Authentication Session
     * @param {CardData} cardData - containing card info
     */
     init3DSecAuthSession   (cardData : CardData) : void {
        // upcast card Data to Model data
        const modelData = this.forceCast<ModelData>(cardData);
        // Extend card model with browser/screen/window properties 
        modelData.windowWidth = self.innerWidth;
        modelData.windowHeight = self.innerHeight;
        modelData.timeZoneOffsetMinutes = new Date().getTimezoneOffset();
        modelData.colorDepth = screen.colorDepth;
        modelData.browserScreenWidth = screen.availWidth;
        modelData.browserScreenHeight = screen.availHeight;
        modelData.browserJavaEnabled = false;
        if (cardData.customerEmail) modelData.customerEmail = cardData.customerEmail;
        if (cardData.customerPhone) modelData.customerPhone = cardData.customerPhone;
        var data = JSON.stringify(modelData);
        console.log(data);
      
        const xhr = new XMLHttpRequest();

        xhr.onload = () => {
            // Request finished. Do processing here.
            if (xhr.status < 300) {
            // Success
                var response = JSON.parse(xhr.responseText);
                this.handleInitSessionResponse(response)
            }
            else {
                // Failure
                this.handleError("Status:" + xhr.status + "; " + xhr.responseText,"init3DSecAuthSession Error");
            }
        };
          
        xhr.ontimeout = () => {
            // XMLHttpRequest timed out
            this.handleError("Http post request timed out","init3DSecAuthSession");
        };

        xhr.open("POST", this.apiBaseUrl + "/threeDsSessions/init", true);
        xhr.timeout = 20000; // 20 sec. (time in milliseconds)
        xhr.setRequestHeader("Content-type", "application/json");
        xhr.setRequestHeader("Authorization", "Bearer " + this.sessionToken);
        xhr.send(data);
    }

    /**
     * Handle response for init session
     * 
     * @param {HostMessage} message - message received from server
     */
     handleInitSessionResponse (message : HostMessage): void {
        console.log('init: Recieved ' + message.clientMessage.messageType + ' message from server');
        switch (message.clientMessage.messageType) {
            case 'TechnicalProblem':
            case 'Rejected':
            case 'Authenticated':
            case 'CardNotEnrolled':
                // no more action required;
                this.resolver(message.clientMessage);
                break;

            case 'ChallengeRequired':
                // Store new token
                this.sessionToken = message.token;
                const m = this.forceCast<ClientChallengeRequiredMessage>(message.clientMessage);
                this.performChallenge(m.ascUri);
                break;     

            case 'FingerPrintRequired':
                // Store new token
                this.sessionToken = message.token;

                console.log("Creating a fingerprint iFrame");
                var ifrm = document.createElement("iframe");
                ifrm.style.display = "none";
                document.body.appendChild(ifrm);

                // inject form into iFrame, then perform a submit operation
                const fprm = this.forceCast<ClientFingerPrintRequiredMessage>(message.clientMessage);
                var body =
                "<body onload='document.getElementById(\"frmRedirect\").submit();'>" +
                "<form method='post' id='frmRedirect' action='" + fprm.fingerPrintUri + "'><input type='hidden' name='threeDSMethodData' value='" + fprm.fingerPrintData + "' /><input type='submit'/></form>" +
                "</body>";
 
                var doc = ifrm.contentWindow?.document;
                doc?.open();
                doc?.write(body);
                doc?.close();


                // Set timer on overall Fingerprint flow
                this.timerPollingHandle = window.setTimeout(()=>{
                    // try obtaining fingerprint status even if notification has not been received
                    this.stopPollingResults();
                    // Clear fingerprint iframe
                    ifrm.parentNode?.removeChild(ifrm);
                    // Resolve with message
                    this.handleError("Timeout on receiving  response after creating a fingerprint iFrame","fingerprintrequired")
                }, 20000);
                
                // Start polling
                this.startPollResults("fingerPrintStatus?isNotified=false");    
                break;
            default:
                const err = 'Received unknown message - ' + message.clientMessage.messageType;
                console.error(err);
                throw new Error(err);
        }
    }

    /**
     * Start polling a 3DS result
     * @param {string} path - rightmost part of the path
     */
     startPollResults(path: string): void {
        this.stopPollingSemaphore = false;

        // Start the polling interval
        this.backgroundPollingInterval = window.setTimeout( (path: string) => {
            // perform a get operation
            console.log('Trying to obtain status from ' + path);
            // This ajax call will retrieve 3DS authentication status
            // Send a request to get results
            const xhr = new XMLHttpRequest();

            // Set handlers for different responses; Each will set the resolver

            // response handler
            xhr.onload = () => { 
                if (xhr.status < 300) {
                    console.log(path + ' inquery returned successfuly');
                    const message: ClientMessage = JSON.parse(xhr.responseText);
                    console.log('Received message: ' + message.messageType);

                    // if result not available yet, proceed by resetting a deferred function.
                    if (message.messageType === "NoResponse") {
                        if (!this.stopPollingSemaphore) {
                            console.log('Response not ready yet; Keep on polling');
                            this.startPollResults(path);
                        } else {
                            console.log('Response not ready yet; Stop polling was signalled => polling stopped');
                            this.handleError("Timeout while Polling results","Polling timeout");
                        }
                        return;
                    }
                    // if result available, stop polling (do not set the timer again) and handle the message
                    console.log('Response received; Stop polling and handle message');
                    this.stopPollingResults();
                    if (message.messageType === "ChallengeRequired") {
                        // challenge required
                        const challemgeMessage = this.forceCast<ClientChallengeRequiredMessage>(message);
                        this.performChallenge(challemgeMessage.ascUri);
                    }
                    else {
                        // final status
                        const authenticatedMessage = this.forceCast<ClientAuthenticatedMessage>(message);
                        this.handleServerResultMessage(authenticatedMessage);
                    }
                    return;
                }
                else {
                    this.handleError(xhr.responseText,"Polling error");
                }
            };

            // Timeout handler
            xhr.ontimeout = (e) => {
                // XMLHttpRequest timed out.
                this.handleError("Http get request timed out","Polling")
            };
            xhr.open("GET", this.apiBaseUrl + "/threeDsSessions/" + path, true);
            xhr.timeout = 20000; // 20 sec
            xhr.setRequestHeader("Authorization", "Bearer " + this.sessionToken);
            xhr.send();
        }, 2000, path); // every 2 seconds
    }

    /**
     * Perform challenge. Can be called either after session initiation or after a fingerprint test
     * @param {string} ascUri - the URI to use for iFrame src tag.
     */
     performChallenge(ascUri: string) : void {
        // Set timer for Challenge polling
        this.timerPollingHandle = window.setTimeout(() => {
            // Hide challenge iframe
            this.iFrame.style.display = "none";

            this.handleError("Timeout on receiving response after creating a challenge iFrame","challengerequired");

        }, 600000 /* 10 min. */);

        // Open iframe with source to challenge over the entire existing window
        console.log("Creating a challenge iFrame");

        this.iFrame.style.display = "inline";
        this.iFrame.setAttribute("src", ascUri);

        // Start polling challenge results
        this.startPollResults("challengeStatus");
    }

    /**
     * Stop polling for results
     */
    stopPollingResults(): void {
        console.log('Signalling a Stop polling and stopping a possible pending poll');
        this.stopPollingSemaphore = true;
        window.clearTimeout(this.backgroundPollingInterval);
        window.clearTimeout(this.timerPollingHandle);
    }

    /**
     * Handle Server result Message
     * 
     * @param {*} message - message to be sent back to caller
     */
     handleServerResultMessage(message: ClientMessage): void {
        console.log('3DS: Received ' + message.messageType + ' message from server');
        // Hide challenge iframe
        this.iFrame.style.display = "none";

        this.resolver(message);
    }

    /**
     * Error reporting
     * @param {string} reason
     * @param {string} ref
     */
    handleError(reason: string, ref: string) : void {
        console.error(ref + ': ' + reason);   
        const errMessage = new ClientTechnicalProblemDetectedMessage();
        errMessage.messageType  = "TechnicalProblem";
        errMessage.reason = reason;
        errMessage.reference = ref;

        this.resolver(errMessage);
    }

    /** 
     * Cast an object to a given type 
     */
    forceCast<T>(input: any): T { 
        // @ts-ignore <-- forces TS compiler to compile this as-is
        return input;
    }
}